import re
import logging
from lxml import etree
from .tree import Tree


logger = logging.getLogger(__name__)


def to_xml(nbest_trees, tagged_doc):
    candc_node = etree.Element('candc')
    for i, (nbest, tokens) in enumerate(zip(nbest_trees, tagged_doc), 1):
        for j, (tree, _) in enumerate(nbest, 1):
            out = tree.xml(tokens)
            out.set('sentence', str(i))
            out.set('id', str(j))
            candc_node.append(out)
    return candc_node


def to_prolog(nbest_trees, tagged_doc):
    out = (':- op(601, xfx, (/)).\n'
           ':- op(601, xfx, (\)).\n'
           ':- multifile ccg/2, id/2.\n'
           ':- discontiguous ccg/2, id/2.\n\n')
    for i, (nbest, tokens) in enumerate(zip(nbest_trees, tagged_doc), 1):
        for token in tokens:
            if "'" in token.lemma:
                token.lemma = token.lemma.replace("'", "\\'")
        for t, _ in nbest:
            out += t.prolog().format(i, *tokens) + '\n'
    return out


MATHML_SUBTREE_NONTERMINAL = '''\
<mrow>
  <mfrac {3} linethickness='2px'>
    <mrow>{0}</mrow>
    <mstyle mathcolor='Red'>{1}</mstyle>
  </mfrac>
  <mtext mathsize='0.8' mathcolor='Black'>{2}</mtext>
</mrow>
'''

MATHML_SUBTREE_TERMINAL = '''\
<mrow>
  <mfrac linethickness='2px'>
    <mtext mathsize='1.0' mathcolor='Black'>{0}</mtext>
    <mstyle mathcolor='Red'>{1}</mstyle>
  </mfrac>
  <mtext mathsize='0.8' mathcolor='Black'>lex</mtext>
</mrow>
'''

MATHML_MAIN = '''\
<!doctype html>
<html lang='en'>
<head>
  <meta charset='UTF-8'>
  <style>
    body {{
      font-size: 1em;
    }}
  </style>
  <script type="text/javascript"
     src="http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML">
  </script>
</head>
<body>
  {0}
</body>
</html>
'''


def mathml_subtree(tree, bgcolor=None):
    bgcolor = f'mathbackground={bgcolor}' if bgcolor else ''
    cat_str = mathml_cat(str(tree.cat))
    if not tree.is_leaf:
        children_str = ''.join(map(mathml_subtree, tree.children))
        return MATHML_SUBTREE_NONTERMINAL.format(
            children_str, cat_str, tree.op_string, bgcolor)
    else:
        return MATHML_SUBTREE_TERMINAL.format(
            tree.word, cat_str)


def mathml_cat(cat):
    cats_feats = re.findall(r'([\w\\/()]+)(\[.+?\])*', cat)
    mathml_str = ''
    for cat, feat in cats_feats:
        cat_mathml = f'''\
<mi mathvariant='italic'
  mathsize='1.0' mathcolor='Red'>{cat}</mi>'''
        if feat != '':
            mathml_str += f'''\
<msub>{cat_mathml}
  <mrow>
  <mi mathvariant='italic'
    mathsize='0.8' mathcolor='Purple'>{feat}</mi>
  </mrow>
</msub>'''
        else:
            mathml_str += cat_mathml
    return mathml_str


def to_mathml(trees):
    result = ''
    for i, nbest in enumerate(trees):
        words = nbest[0][0].word if isinstance(nbest[0], tuple) else nbest[0].word
        result += f'<p>ID={i}: {words}</p>'
        for tree in nbest:
            if isinstance(tree, tuple):
                tree, prob = tree
                result += f'<p>Log prob={prob:.5e}</p>'
            tree_str = tree if isinstance(tree, str) else mathml_subtree(tree)
            result += f'<math xmlns="http://www.w3.org/1998/Math/MathML">{tree_str}</math>'
    return MATHML_MAIN.format(result)


class ConvertToJiggXML(object):
    def __init__(self, sid: int):
        self.sid = sid
        self._spid = -1
        self.processed = 0

    @property
    def spid(self) -> int:
        self._spid += 1
        return self._spid

    def process(self, tree: Tree, score: float = None):
        def traverse(node: Tree):
            id = f's{self.sid}_sp{self.spid}'
            xml_node = etree.SubElement(res, 'span')
            xml_node.set('category', str(node.cat.multi_valued))
            xml_node.set('begin', str(node.start_of_span))
            xml_node.set('end', str(node.start_of_span+len(node)))
            xml_node.set('id', id)
            if node.is_leaf:
                xml_node.set('terminal', f't{self.sid}_{node.head_id}')
            else:
                childid = traverse(node.left_child)
                if not node.is_unary:
                    tmp = traverse(node.right_child)
                    childid += ' ' + tmp
                xml_node.set('child', childid)
                xml_node.set('rule', node.op_string)
            return id

        res = etree.Element('ccg')
        res.set('id', f's{self.sid}_ccg{self.processed}')
        id = traverse(tree)
        res.set('root', str(id))
        res[0].set('root', 'true')
        if score is not None:
            res.set('score', str(score))
        self.processed += 1
        return res


def to_jigg_xml(trees, tagged_doc):
    root_node = etree.Element('root')
    document_node = etree.SubElement(root_node, 'document')
    sentences_node = etree.SubElement(document_node, 'sentences')
    for i, (parsed, tagged) in enumerate(zip(trees, tagged_doc)):
        sentence_node = etree.SubElement(sentences_node, 'sentence')
        tokens_node = etree.SubElement(sentence_node, 'tokens')
        cats = [leaf.cat for leaf in parsed[0][0].leaves]
        assert len(cats) == len(tagged)
        for j, (token, cat) in enumerate(zip(tagged, cats)):
            token_node = etree.SubElement(tokens_node, 'token')
            token_node.set('start', str(j))
            token_node.set('cat', str(cat))
            token_node.set('id', f't{i}_{j}')
            if 'word' in token:
                token['surf'] = token.pop('word')
            if 'lemma' in token:
                token['base'] = token.pop('lemma')
            for k, v in token.items():
                token_node.set(k, v)
        converter = ConvertToJiggXML(i)
        for tree, score in parsed:
            sentence_node.append(converter.process(tree, score))
    return root_node





from __future__ import print_function

from libc.stdlib cimport malloc, free

cimport numpy as np
import numpy as np
import sys
import re
import tarfile
import os
import json
import chainer
import logging
from pathlib import Path
from cython.parallel cimport prange
from libc.stdio cimport fprintf, stderr
from .combinator cimport combinator_list_to_vector
from .combinator import en_default_binary_rules, ja_default_binary_rules
from .utils cimport *
from .utils import maybe_split_and_join, denormalize
from .cat cimport Category


logger = logging.getLogger(__name__)


cdef PartialConstraints build_nonterminal_constraints(
        list py_constraints, const unordered_map[Cat, unordered_set[Cat]]& unary_rules):
    cdef PartialConstraints c_constraints = PartialConstraints(unary_rules)
    cdef Category cat
    cdef int start_of_span, span_length
    for cat, start_of_span, span_length in py_constraints:
        if cat is not None:
            c_constraints.Add(cat.cat_, start_of_span, span_length)
        else:
            c_constraints.Add(start_of_span, span_length)
    return c_constraints


def build_terminal_constraints(list constraints, tag_probs, tag_list):
    pseudo_neginf = -10e10
    tag_dict = {tag: i for i, tag in enumerate(tag_list)}
    sent_len, old_tag_size = tag_probs.shape
    if len (tag_list) != old_tag_size:
        assert len(tag_list) > old_tag_size
        new_tag_probs = np.full((sent_len, len(tag_list)), pseudo_neginf, 'f')
        new_tag_probs[:sent_len, :old_tag_size] = tag_probs
    else:
        new_tag_probs = tag_probs
    for cat, i in constraints:
        cat_index = tag_dict[cat]
        new_tag_probs[i, :] = pseudo_neginf
        new_tag_probs[i, cat_index] = 0
    return new_tag_probs


from itertools import islice
def lazy_groups_of(iterator, group_size):
    return iter(lambda: list(islice(iterator, 0, group_size)), [])


class AllennlpSupertagger(object):
    def __init__(self, predictor):
        self.predictor = predictor
        self.dataset_reader = predictor._dataset_reader

    def predict_doc(self, splitted, batchsize=32, gpu=None):
        instances = iter(self.dataset_reader.text_to_instance(' '.join(sentence))
                         for sentence in splitted)

        categories = None
        probs = []
        import json
        for batch in lazy_groups_of(instances, batchsize):
            for json_dict in self.predictor.predict_batch_instance(batch):
                if categories is None:
                    categories = [Category.parse(cat) for cat in json_dict['categories']]
                dep = np.array(json_dict['heads']).reshape(json_dict['heads_shape']).astype(np.float32)
                tag = np.array(json_dict['head_tags']).reshape(json_dict['head_tags_shape']).astype(np.float32)
                probs.append((tag, dep))
        return probs, categories


ctypedef ApplyBinaryRules (*MakeApplyBinaryRules)(const vector[Op]&)


cdef class EnglishCCGParser:
    cdef ApplyUnaryRules apply_unary_rules_
    cdef MakeApplyBinaryRules make_apply_binary_rules_

    cdef object binary_rules
    cdef object possible_root_cats
    cdef object category_dict
    cdef object unary_rules
    cdef object seen_rules
    cdef object unary_penalty
    cdef object beta
    cdef object use_beta
    cdef object use_category_dict
    cdef object use_seen_rules
    cdef object pruning_size
    cdef object nbest
    cdef object max_length
    cdef object max_steps
    cdef object tagger
    cdef object gpu
    cdef bytes lang

    def __init__(self,
                 category_dict,
                 unary_rules,
                 seen_rules,
                 binary_rules=None,
                 unary_penalty=0.1,
                 beta=0.00001,
                 use_beta=True,
                 use_category_dict=True,
                 use_seen_rules=True,
                 pruning_size=50,
                 nbest=1,
                 possible_root_cats=None,
                 max_length=250,
                 max_steps=100000,
                 gpu=-1):
        self.apply_unary_rules_ = EnApplyUnaryRules
        self.make_apply_binary_rules_ = MakeEnApplyBinaryRules
        self.binary_rules = binary_rules or en_default_binary_rules
        if possible_root_cats is None:
            possible_root_cats = ['S[dcl]', 'S[wq]', 'S[q]', 'S[qem]', 'NP']
        self.possible_root_cats = [Category.parse(cat) if not isinstance(cat, Category) else cat
                                   for cat in possible_root_cats]
        self.category_dict = category_dict
        self.unary_rules = unary_rules
        self.seen_rules = seen_rules
        self.unary_penalty = unary_penalty
        self.beta = beta
        self.use_beta = use_beta
        self.use_category_dict = use_category_dict
        self.use_seen_rules = use_seen_rules
        self.pruning_size = pruning_size
        self.nbest = nbest
        self.max_length = max_length
        self.max_steps = max_steps
        self.tagger = None
        self.gpu = gpu
        self.lang = b'en'

    def load_default_tagger(self, model_path):
        logger.info(f'loading default supertagger at {model_path}')
        if os.path.isdir(model_path) and \
            os.path.exists(os.path.join(model_path, 'tagger_model')):
            self.load_chainer_tagger(model_path)
        else:
            self.load_allennlp_tagger(model_path)

    def load_chainer_tagger(self, dirname):
        from depccg.lstm_parser_bi_fast import FastBiaffineLSTMParser
        from depccg.ja_lstm_parser_bi import BiaffineJaLSTMParser
        model_file = os.path.join(dirname, 'tagger_model')
        def_file = os.path.join(dirname, 'tagger_defs.txt')
        assert os.path.exists(model_file) and os.path.exists(def_file), \
            (f'Failed in initialization. Directory "{dirname}" must contain both'
             '"tagger_model" and "tagger_defs.txt" files')
        with open(def_file) as f:
            self.tagger = eval(json.load(f)['model'])(dirname)
        logger.info(f'initializing supertagger with parameters at {model_file}')
        chainer.serializers.load_npz(model_file, self.tagger)
        if self.gpu >= 0:
            logger.info(f'sending the supertagger to gpu: {self.gpu}')
            self.tagger.to_gpu(self.gpu)

    def load_allennlp_tagger(self, model_path):
        from allennlp.models.archival import load_archive
        from depccg.models.my_allennlp.models.supertagger import Supertagger
        from depccg.models.my_allennlp.dataset.supertagging_dataset import SupertaggingDatasetReader
        from depccg.models.my_allennlp.dataset.supertagging_dataset import TritrainSupertaggingDatasetReader
        from depccg.models.my_allennlp.predictor.supertagger_predictor import SupertaggerPredictor
        if self.gpu >= 0:
            logger.info(f'sending the supertagger to gpu: {self.gpu}')
        archive = load_archive(model_path, cuda_device=self.gpu)
        predictor = SupertaggerPredictor.from_archive(archive, 'supertagger-predictor')
        self.tagger = AllennlpSupertagger(predictor)

    @classmethod
    def from_dir(cls, dirname, load_tagger=False, **kwargs):
        logger.info(f'loading parser from {dirname}')
        args = []
        for file in ['unary_rules.txt', 'cat_dict.txt', 'seen_rules.txt']:
            arg = os.path.join(dirname, file)
            if os.path.exists(arg):
                args.append(arg)
            else:
                args.append(None)
        if load_tagger:
            kwargs['tagger_model'] = dirname
        return cls.from_files(*args, **kwargs)

    @classmethod
    def from_files(cls, unary_rules=None, category_dict=None, seen_rules=None, tagger_model=None, **kwargs):
        files = [file for file in [unary_rules, category_dict, seen_rules, tagger_model]
                 if file is not None]
        logger.info(f'loading parser from files: {files}')
        unary_rules = read_unary_rules(unary_rules) if unary_rules else []
        category_dict = read_cat_dict(category_dict) if category_dict else {}
        seen_rules = read_seen_rules(seen_rules, cls.preprocess_seen_rules) if seen_rules else []
        parser = cls(category_dict, unary_rules, seen_rules, **kwargs)
        if tagger_model:
            parser.load_default_tagger(tagger_model)
        return parser

    @classmethod
    def from_json(cls, json_input, tagger_model=None, **kwargs):
        if isinstance(json_input, (str, Path)):
            logger.info(f'loading parser from json file: {json_input}')
            json_input = json.load(open(json_input))
        else:
            assert isinstance(json_input, dict), \
                'the input to from_json must be either a dict object or filename stirng'
        unary_rules = [(Category.parse(c1), Category.parse(c2)) for c1, c2 in json_input['unary_rules']]
        category_dict = {}
        for word, cats in json_input['cat_dict'].items():
            category_dict[word] = [Category.parse(cat) for cat in cats]

        seen_rules = []
        for c1, c2 in json_input['seen_rules']:
            c1 = cls.preprocess_seen_rules(Category.parse(c1))
            c2 = cls.preprocess_seen_rules(Category.parse(c2))
            seen_rules.append((c1, c2))
        parser = cls(category_dict, unary_rules, seen_rules, **kwargs)
        if tagger_model:
            parser.load_default_tagger(tagger_model)
        return parser

    @classmethod
    def preprocess_seen_rules(cls, cat):
        return cat.strip_feat('[X]').strip_feat('[nb]')

    def parse_doc(self, sents, probs=None, tag_list=None, constraints=None, batchsize=16):
        splitted, sents = zip(*map(maybe_split_and_join, sents))
        logger.info('start tagging sentences')
        if probs is None:
            assert self.tagger is not None, 'default supertagger is not loaded.'
            probs, tag_list = self.tagger.predict_doc(
                    splitted, batchsize=batchsize, gpu=self.gpu)
        else:
            assert tag_list is not None
        logger.info('done tagging sentences')
        res = self._parse_doc_tag_and_dep(list(sents),
                                          list(probs),
                                          tag_list,
                                          constraints=constraints)
        return res

    def parse_json(self, json_input, batchsize=16):
        if isinstance(json_input, str):
            json_input = [json.loads(line.strip()) for line in open(json_input)]
        categories = None
        sents = []
        probs = []
        constraints = []
        unprocessed = {}
        for i, json_dict in enumerate(json_input):
            if categories is None:
                categories = [Category.parse(cat) for cat in json_dict['categories']]

            words = [denormalize(word) for word in json_dict['words'].split(' ')]
            sent = ' '.join(words)
            dep = json_dict.get('heads', None)
            tag = json_dict.get('head_tags', None)

            if dep is None and tag is None:
                def process_fun(_, new_tag_and_dep):
                    return new_tag_and_dep
                tag_and_dep = None
            elif dep is None:
                def process_fun(tag_and_dep, new_tag_and_dep):
                    tag, _ = tag_and_dep
                    _, new_dep = new_tag_and_dep
                    return tag, new_dep
                tag = np.array(tag).reshape(json_dict['head_tags_shape']).astype(np.float32)
                tag_and_dep = (tag, None)
            elif tag is None:
                raise NotImplementedError('not supported.')
            else:
                process_fun = None
                dep = np.array(dep).reshape(json_dict['heads_shape']).astype(np.float32)
                tag = np.array(tag).reshape(json_dict['head_tags_shape']).astype(np.float32)
                tag_and_dep = (tag, dep)

            if process_fun is not None:
                unprocessed[i] = (process_fun, words)

            constraints.append(json_dict.get('constraints', []))
            sents.append(sent)
            probs.append(tag_and_dep)

        if len(unprocessed) > 0:
            logger.info('Detected that some examples lack either p_tag and p_dep. '
                        'Assigning them using default tagger.')
            assert self.tagger is not None, 'default supertagger is not loaded.'
            _, splitted = zip(*unprocessed.values())
            unprocessed_probs, new_categories = self.tagger.predict_doc(
                    splitted, batchsize=batchsize, gpu=self.gpu)
            categories = new_categories
            for (i, (process_fun, _)), new_tag_and_dep in zip(sorted(unprocessed.items()), unprocessed_probs):
                probs[i] = process_fun(probs[i], new_tag_and_dep)
            assert all(tag is not None and dep is not None for tag, dep in probs)

        if all(len(constraint) == 0 for constraint in constraints):
            constraints = None
        else:
            constraints = [[(Category.parse(cat), i, j) for cat, i, j in cxs]
                           for cxs in constraints]
        res = self._parse_doc_tag_and_dep(sents,
                                          probs,
                                          categories,
                                          constraints=constraints)
        return res

    def _parse_doc_tag_and_dep(self, sents, probs, py_tag_list, constraints=None):
        logger.info(f'unary penalty = {self.unary_penalty}')
        logger.info(f'beta value = {self.beta} (use beta = {self.use_beta})')
        logger.info(f'pruning size = {self.pruning_size}')
        logger.info(f'N best = {self.nbest}')
        logger.info(f'use category dictionary = {self.use_category_dict}')
        logger.info(f'use seen rules = {self.use_seen_rules}')
        logger.info(f'allow at the root of a tree only categories in {self.possible_root_cats}'),
        logger.info(f'give up sentences that contain > {self.max_length} words')
        logger.info(f'combinators: {self.binary_rules}')

        py_tag_list = [Category.parse(cat) if isinstance(cat, str) else cat
                       for cat in py_tag_list]
        cdef int doc_size = len(sents), i, j
        cdef np.ndarray[float, ndim=2, mode='c'] cat_scores, dep_scores
        cdef vector[string] csents = vector[string](doc_size)
        cdef float **tags = <float**>malloc(doc_size * sizeof(float*))
        cdef float **deps = <float**>malloc(doc_size * sizeof(float*))

        cdef vector[Cat] tag_list_ = cat_list_to_vector(list(py_tag_list))
        cdef unordered_map[string, unordered_set[Cat]] category_dict_ = \
            convert_cat_dict(self.category_dict) if self.use_category_dict \
                else unordered_map[string, unordered_set[Cat]]()
        cdef float unary_penalty_ = self.unary_penalty
        cdef float beta_ = self.beta
        cdef bool use_beta = self.use_beta
        cdef unsigned pruning_size_ = self.pruning_size
        cdef unsigned nbest_ = self.nbest
        cdef vector[Op] binary_rules_ = combinator_list_to_vector(self.binary_rules)
        cdef unordered_set[Cat] possible_root_cats_ = cat_list_to_unordered_set(self.possible_root_cats)
        cdef unordered_map[Cat, unordered_set[Cat]] unary_rules_ = convert_unary_rules(self.unary_rules)
        cdef unordered_set[CatPair] seen_rules_ = \
            convert_seen_rules(self.seen_rules) if self.use_seen_rules else unordered_set[CatPair]()
        cdef ApplyBinaryRules default_apply_binary_rules = self.make_apply_binary_rules_(binary_rules_)
        cdef ApplyUnaryRules apply_unary_rules_ = self.apply_unary_rules_
        cdef unsigned max_length_ = self.max_length
        cdef unsigned max_steps_ = self.max_steps

        cdef vector[ApplyBinaryRules] apply_binary_rules = \
                vector[ApplyBinaryRules](doc_size, default_apply_binary_rules)
        cdef vector[PartialConstraints] c_constraints

        if constraints is not None:
            assert len(constraints) == doc_size
            logger.info('loading partial constraints')
            c_constraints = vector[PartialConstraints](doc_size)
            new_probs = []

            # update py_tag_list and tag_list_ if there are unseen supertags in constraints
            cats_terminal_constraints = [cx[0] for constraint in constraints for cx in constraint if len(cx) == 2]
            new_terminal_categories = list({cat for cat in cats_terminal_constraints if cat not in py_tag_list})
            if len (new_terminal_categories) > 0:
                logging.info(f'newly added categories in terminal constraints: {new_terminal_categories}')
                py_tag_list += new_terminal_categories
                tag_list_ = cat_list_to_vector(py_tag_list)

            for i, (constraint, (py_cat_scores, py_dep_scores)) in enumerate(zip(constraints, probs)):
                nonterminal_constraints = [cx for cx in constraint if len(cx) == 3]
                terminal_constraints = [cx for cx in constraint if len(cx) == 2]
                logging.debug(f'non-terminal constraints: {nonterminal_constraints}')
                logging.debug(f'terminal constraints: {terminal_constraints}')
                c_constraints[i] = build_nonterminal_constraints(nonterminal_constraints, unary_rules_)
                py_cat_scores = build_terminal_constraints(terminal_constraints, py_cat_scores, py_tag_list)
                new_probs.append((py_cat_scores, py_dep_scores))
            probs = new_probs
        else:
            c_constraints = vector[PartialConstraints](doc_size, PartialConstraints())

        cdef int tag_size = tag_list_.size()
        for i, (py_cat_scores, py_dep_scores) in enumerate(probs):
            sent_size = len(sents[i].split(' '))
            if (sent_size, tag_size) != py_cat_scores.shape or \
                (sent_size, sent_size + 1) != py_dep_scores.shape:
                raise RuntimeError(
                    'invalid shape of input matrices:\n'
                    f'Expected P_tag: ({sent_size}, {tag_size}), P_dep: ({sent_size}, {sent_size + 1})\n'
                    f'Actual P_tag: {py_cat_scores.shape}, P_dep: {py_dep_scores.shape}')
            cat_scores = py_cat_scores
            dep_scores = py_dep_scores
            csents[i] = sents[i].encode('utf-8')
            tags[i] = &cat_scores[0, 0]
            deps[i] = &dep_scores[0, 0]

        logger.info('start A* parsing')
        cdef vector[vector[ScoredNode]] cres = ParseSentences(
                   csents,
                   tags,
                   deps,
                   category_dict_,
                   tag_list_,
                   unary_penalty_,
                   beta_,
                   use_beta,
                   pruning_size_,
                   nbest_,
                   possible_root_cats_,
                   unary_rules_,
                   seen_rules_,
                   apply_binary_rules,
                   apply_unary_rules_,
                   c_constraints,
                   max_length_,
                   max_steps_)
        logger.info('done A* parsing')
        cdef list tmp, res = []
        cdef Tree parse
        for i in range(len(sents)):
            tmp = []
            for j in range(min(nbest_, cres[i].size())):
                tmp.append((Tree.from_ptr(cres[i][j].first, self.lang), cres[i][j].second))
            res.append(tmp)
        free(tags)
        free(deps)
        return res


cdef class JapaneseCCGParser(EnglishCCGParser):
    def __init__(self,
                 category_dict,
                 unary_rules,
                 seen_rules,
                 binary_rules=None,
                 unary_penalty=0.1,
                 beta=0.00001,
                 use_beta=True,
                 use_category_dict=True,
                 use_seen_rules=True,
                 pruning_size=50,
                 nbest=1,
                 possible_root_cats=None,
                 max_length=250,
                 max_steps=100000,
                 gpu=-1):

        binary_rules = binary_rules or ja_default_binary_rules
        if possible_root_cats is None:
            possible_root_cats = [
                'NP[case=nc,mod=nm,fin=f]',
                'NP[case=nc,mod=nm,fin=t]',
                'S[mod=nm,form=attr,fin=t]',
                'S[mod=nm,form=base,fin=f]',
                'S[mod=nm,form=base,fin=t]',
                'S[mod=nm,form=cont,fin=f]',
                'S[mod=nm,form=cont,fin=t]',
                'S[mod=nm,form=da,fin=f]',
                'S[mod=nm,form=da,fin=t]',
                'S[mod=nm,form=hyp,fin=t]',
                'S[mod=nm,form=imp,fin=f]',
                'S[mod=nm,form=imp,fin=t]',
                'S[mod=nm,form=r,fin=t]',
                'S[mod=nm,form=s,fin=t]',
                'S[mod=nm,form=stem,fin=f]',
                'S[mod=nm,form=stem,fin=t]'
            ]

        self.apply_unary_rules_ = DefaultApplyUnaryRules
        self.make_apply_binary_rules_ = MakeDefaultApplyBinaryRules
        self.possible_root_cats = [Category.parse(cat) if not isinstance(cat, Category) else cat
                                   for cat in possible_root_cats]
        self.category_dict = category_dict
        self.unary_rules = unary_rules
        self.seen_rules = seen_rules
        self.binary_rules = binary_rules
        self.unary_penalty = unary_penalty
        self.beta = beta
        self.use_beta = use_beta
        self.use_category_dict = use_category_dict
        self.use_seen_rules = use_seen_rules
        self.pruning_size = pruning_size
        self.nbest = nbest
        self.max_length = max_length
        self.max_steps = max_steps
        self.tagger = None
        self.gpu = gpu
        self.lang = b'ja'

    @classmethod
    def preprocess_seen_rules(cls, cat):
        return cat




from libcpp.string cimport string
from libcpp.pair cimport pair
from cython.operator cimport dereference as deref


cdef class Category:
    def __cinit__(self):
        pass

    def __str__(self):
        return self.cat_.ToStr().decode('utf-8')

    def __repr__(self):
        return self.cat_.ToStr().decode('utf-8')

    @staticmethod
    def parse(cat):
        if not isinstance(cat, bytes):
            cat = cat.encode('utf-8')
        c = Category()
        c.cat_ = CCategory.Parse(cat)
        return c

    @staticmethod
    cdef Category from_ptr(Cat cat):
        c = Category()
        c.cat_ = cat
        return c

    cdef bool equals_to(self, Category other):
        return deref(self.cat_) == deref(other.cat_)

    def __eq__(self, other):
       if isinstance(other, Category):
           return self.equals_to(other)
       else:
           return False

    def __hash__(self):
        return self.cat_.GetId()

    property multi_valued:
        def __get__(self):
            return Category.from_ptr(self.cat_.ToMultiValue())

    property base:
        def __get__(self):
            return self.cat_.ToStrWithoutFeat().decode('utf-8')

    property left:
        def __get__(self):
            assert self.is_functor, \
                "Error {} is not functor type.".format(str(self))
            return Category.from_ptr(self.cat_.GetLeft())

    property right:
        def __get__(self):
            assert self.is_functor, \
                "Error {} is not functor type.".format(str(self))
            return Category.from_ptr(self.cat_.GetRight())

    property is_modifier:
        def __get__(self):
            return self.cat_.IsModifier()

    property is_modifier_without_feat:
        def __get__(self):
            return self.cat_.IsModifierWithoutFeat()

    property is_type_raised:
        def __get__(self):
            return self.cat_.IsTypeRaised()

    property is_type_raised_without_feat:
        def __get__(self):
            return self.cat_.IsTypeRaisedWithoutFeat()

    property is_forward_type_raised:
        def __get__(self):
            return self.cat_.IsForwardTypeRaised()

    property is_backward_type_raised:
        def __get__(self):
            return self.cat_.IsBackwardTypeRaised()

    property is_functor:
        def __get__(self):
            return self.cat_.IsFunctor()

    property is_punct:
        def __get__(self):
            return self.cat_.IsPunct()

    property is_NorNP:
        def __get__(self):
            return self.cat_.IsNorNP()

    def is_function_into(self, cat):
        return self._is_function_into(cat)

    cdef bint _is_function_into(self, Category cat):
        return self.cat_.IsFunctionInto(cat.cat_)

    property n_args:
        def __get__(self):
            return self.cat_.NArgs()

    def matches(self, other):
        return self._matches(other)

    cdef bint _matches(self, Category other):
        return self.cat_.Matches(other.cat_)

    def arg(self, i):
        return Category.from_ptr(self.cat_.Arg(i))

    property slash:
        def __get__(self):
            assert self.is_functor, 'Category "{}" is not a functor type.'.format(str(self))
            return self.cat_.GetSlash().ToStr().decode('utf-8')

    property features:
        def __get__(self):
            cdef unordered_map[string, string] c_features
            cdef pair[string, string] tmp
            cdef str key, val
            if self.is_functor:
                return {}
            else:
                c_features = self.cat_.GetFeat().Values()
                res = {}
                for tmp in c_features:
                    key = tmp.first.decode('utf-8')
                    val = tmp.second.decode('utf-8')
                    res[key] = val
                return res


    cpdef strip_feat(self, feat):
        assert feat.startswith('[') and feat.endswith(']'), \
            'please enclose a feature with [] as in Category.parse(\'S[dcl]\\NP\').strip_feat(\'[dcl]\')'
        cdef string c_feat = feat.encode('utf-8')
        return Category.from_ptr(self.cat_.StripFeat(c_feat))

    def json(self):
        def rec(node):
            if node.is_functor:
                return {
                    'slash': node.slash,
                    'left': rec(node.left),
                    'right': rec(node.right)
                }
            else:
                feature = node.features
                return {
                    'base': node.base,
                    'feature': feature if len(feature) > 0 else None
                }
        return rec(self)





import sys
import tarfile
from urllib.request import urlretrieve
import logging
import time
from pathlib import Path


logger = logging.getLogger(__name__)

MODELS = {
    'en': ('tri_headfirst', 'http://cl.naist.jp/~masashi-y/resources/depccg/en_hf_tri.tar.gz'),
    'ja': ('ja_headfinal', 'http://cl.naist.jp/~masashi-y/resources/depccg/ja_hf_ccgbank.tar.gz')
}

MODEL_DIRECTORY = Path(__file__).parent / 'models'

SEMANTIC_TEMPLATES = {
    'en': MODEL_DIRECTORY / 'semantic_templates_en_event.yaml',
    'ja': MODEL_DIRECTORY / 'semantic_templates_ja_event.yaml'
}

CONFIGS = {
    'en': MODEL_DIRECTORY / 'config_en.json',
    'ja': MODEL_DIRECTORY / 'config_ja.json'
}

def reporthook(count, block_size, total_size):
    global start_time
    if count == 0:
        start_time = time.time()
        return
    duration = time.time() - start_time
    progress_size = int(count * block_size)
    speed = int(progress_size / (1024 * duration))
    percent = min(int(count * block_size * 100 / total_size), 100)
    sys.stdout.write("\r...%d%%, %d MB, %d KB/s, %d seconds passed" %
                     (percent, progress_size / (1024 * 1024), speed, duration))
    sys.stdout.flush()


def download(model_name):
    basename, url = MODELS[model_name]
    logging.info(f'start downloading from {url}')
    filename = (MODEL_DIRECTORY / basename).with_suffix('.tar.gz')
    urlretrieve(url, filename, reporthook)
    logging.info(f'extracting files')
    tf = tarfile.open(filename)
    tf.extractall(MODEL_DIRECTORY)
    logging.info(f'finished')


def load_model_directory(model_name):
    basename, url = MODELS[model_name]
    model_dir = MODEL_DIRECTORY / basename
    if not model_dir.exists():
        raise RuntimeError(f'please download the model by doing \'depccg_{model_name} download\'.')
    return model_dir



from libcpp.string cimport string
from cython.operator cimport dereference as deref
from libcpp.pair cimport pair
from libcpp.memory cimport shared_ptr, make_shared
from .cat cimport Cat, Category
from .combinator import unknown_combinator
from lxml import etree
from .token import Token
from depccg.utils import denormalize


## TODO: ugly code
cdef ResolveCombinatorName(const Node* tree, bytes lang):
    cdef string res;
    if lang == b"en":
        res = EnResolveCombinatorName(tree)
    elif lang == b"ja":
        res = JaResolveCombinatorName(tree)
    else:
        res = b"error: " + lang
    return res.decode("utf-8")


unknown_op = unknown_combinator()


class _AutoLineReader(object):
    def __init__(self, line, lang):
        self.line = line
        self.index = 0
        self.word_id = -1
        self.lang = lang

    def next(self):
        end = self.line.find(' ', self.index)
        res = self.line[self.index:end]
        self.index = end + 1
        return res

    def check(self, text, offset=0):
        if self.line[self.index + offset] != text:
            raise RuntimeError(f'failed to parse: {self.line}')

    def peek(self):
        return self.line[self.index]

    def parse(self):
        return self.next_node()

    @property
    def next_node(self):
        if self.line[self.index+2] == 'L':
            return self.parse_leaf
        elif self.line[self.index+2] == 'T':
            return self.parse_tree
        else:
            raise RuntimeError(f'failed to parse: {self.line}')

    def parse_leaf(self):
        self.word_id += 1
        self.check('(')
        self.check('<', 1)
        self.check('L', 2)
        self.next()
        cat = Category.parse(self.next())
        tag1 = self.next()  # modified POS tag
        tag2 = self.next()  # original POS
        word = self.next().replace('\\', '')
        if word == '-LRB-':
            word = "("
        elif word == '-RRB-':
            word = ')'
        dep = self.next()[:-2]
        return Tree.make_terminal(word, cat, self.word_id, self.lang)

    def parse_tree(self):
        self.check('(')
        self.check('<', 1)
        self.check('T', 2)
        self.next()
        cat = Category.parse(self.next())
        left_is_head = self.next() == '0'
        self.next()
        children = []
        while self.peek() != ')':
            children.append(self.next_node())
        self.next()
        if len(children) == 2:
            left, right = children
            return Tree.make_binary(
                cat, left_is_head, left, right, unknown_op, self.lang)
        elif len(children) == 1:
            return Tree.make_unary(cat, children[0], self.lang)
        else:
            raise RuntimeError(f'failed to parse: {self.line}')


cdef class Tree:
    @staticmethod
    cdef Tree from_ptr(NodeType node, lang):
        if isinstance(lang, str):
            lang = lang.encode('utf-8')
        p = Tree()
        p.node_ = node
        p.lang = lang
        return p

    @staticmethod
    def make_terminal(str word, Category cat, int position, lang):
        cdef string c_word = word.encode('utf-8')
        cdef NodeType node = <NodeType>make_shared[Leaf](c_word, cat.cat_, position)
        return Tree.from_ptr(node, lang)

    @staticmethod
    def make_binary(Category cat, bool left_is_head, Tree left, Tree right, Combinator op, lang):
        cdef NodeType node = <NodeType>make_shared[CTree](
            cat.cat_, left_is_head, left.node_, right.node_, op.op_)
        return Tree.from_ptr(node, lang)

    @staticmethod
    def make_unary(Category cat, Tree child, lang):
        cdef NodeType node = <NodeType>make_shared[CTree](cat.cat_, child.node_)
        return Tree.from_ptr(node, lang)

    @staticmethod
    def of_auto(line, lang='en'):
        return _AutoLineReader(line, lang).parse()

    @staticmethod
    def of_nltk_tree(tree, lang='en'):
        position = [-1]
        def rec(node):
            cat = Category.parse(node.label())
            if isinstance(node[0], str):
                word = node[0]
                position[0] += 1
                return Tree.make_terminal(word, cat, position[0], lang)
            else:
                children = [rec(child) for child in node]
                if len(children) == 1:
                    return Tree.make_unary(cat, children[0], lang)
                else:
                    assert len(children) == 2
                    left, right = children
                    return Tree.make_binary(
                        cat, True, left, right, unknown_op, lang)
        return rec(tree)

    def __cinit__(self):
        self.suppress_feat = False

    property cat:
        def __get__(self):
            return Category.from_ptr(deref(self.node_).GetCategory())

    property op_string:
        def __get__(self):
            assert not self.is_leaf, "This node is leaf and does not have combinator!"
            cdef const Node* c_node = &deref(self.node_)
            return ResolveCombinatorName(c_node, self.lang)

    def __len__(self):
        return deref(self.node_).GetLength()

    property children:
        def __get__(self):
            if self.is_leaf:
                return []
            res = [self.left_child]
            if not self.is_unary:
                res.append(self.right_child)
            return res

    property leaves:
        def __get__(self):
            def rec(node):
                if node.is_leaf:
                    res.append(node)
                else:
                    for child in node.children:
                        rec(child)
            res = []
            rec(self)
            return res

    property left_child:
        def __get__(self):
            assert not self.is_leaf, "This node is leaf and does not have any child!"
            return Tree.from_ptr(<NodeType>deref(self.node_).GetLeftChild(), self.lang)

    property right_child:
        def __get__(self):
            assert not self.is_leaf, "This node is leaf and does not have any child!"
            assert not self.is_unary, "This node does not have right child!"
            return Tree.from_ptr(<NodeType>deref(self.node_).GetRightChild(), self.lang)

    property is_leaf:
        def __get__(self):
            return deref(self.node_).IsLeaf()

    property start_of_span:
        def __get__(self):
            return deref(self.node_).GetStartOfSpan()

    property word:
        def __get__(self):
            cdef string res = deref(self.node_).GetWord()
            return res.decode("utf-8")

    property head_id:
        def __get__(self):
            return deref(self.node_).GetHeadId()

    property dependency_length:
        def __get__(self):
            return deref(self.node_).GetDependencyLength()

    property head_is_left:
        def __get__(self):
            return deref(self.node_).HeadIsLeft()

    property is_unary:
        def __get__(self):
            return deref(self.node_).IsUnary()

    property num_descendants:
        def __get__(self):
            return deref(self.node_).NumDescendants()

    property right_num_descendants:
        def __get__(self):
            return deref(self.node_).RightNumDescendants()

    property left_num_descendants:
        def __get__(self):
            return deref(self.node_).LeftNumDescendants()

    def __str__(self):
        return self.auto()

    def __repr__(self):
        return self.auto()

    def ptb(self):
        def rec(node):
            if node.is_leaf:
                cat = node.cat
                word = node.word
                return f'({cat} {word})'
            else:
                cat = node.cat
                children = ' '.join(rec(child) for child in node.children)
                return f'({cat} {children})'
        return f'(ROOT {rec(self)})'

    def nltk_tree(self):
        from nltk.tree import Tree
        def rec(node):
            if node.is_leaf:
                cat = node.cat
                children = [node.word]
            else:
                cat = node.cat
                children = [rec(child) for child in node.children]
            return Tree(str(cat), children)
        return rec(self)

    def auto(self, tokens=None):
        def rec(node):
            if node.is_leaf:
                cat = node.cat
                word = denormalize(node.word)
                pos = poss.pop(0)
                return f'(<L {cat} {pos} {pos} {word} {cat}>)'
            else:
                cat = node.cat
                children = ' '.join(rec(child) for child in node.children)
                num_children = len(node.children)
                head_is_left = 0 if node.head_is_left else 1
                return f'(<T {cat} {head_is_left} {num_children}> {children} )'
        if tokens:
            poss = [token.pos for token in tokens]
        else:
            poss = ['POS' for _ in range(len(self))]
        return rec(self)

    def deriv(self):
        cdef string res = Derivation(self.node_, not self.suppress_feat).Get()
        return res.decode("utf-8")

    def xml(self, tokens=None):
        def rec(node, parent):
            if node.is_leaf:
                leaf_node = etree.SubElement(parent, 'lf')
                start, token = tokens.pop(0)
                leaf_node.set('start', str(start))
                leaf_node.set('span', '1')
                leaf_node.set('cat', str(node.cat))
                for k, v in token.items():
                    leaf_node.set(k, v)
            else:
                rule_node = etree.SubElement(parent, 'rule')
                rule_node.set('type', node.op_string)
                rule_node.set('cat', str(node.cat))
                for child in node.children:
                    rec(child, rule_node)

        if tokens is None:
            tokens = [Token.from_word(word) for word in self.word.split(' ')]
        tokens = list(enumerate(tokens))
        res = etree.Element("ccg")
        rec(self, res)
        return res

    def json(self, tokens=None, full=False):
        def rec(node):
            if node.is_leaf:
                token = tokens.pop(0)
                res = dict(token)
                res['cat'] = node.cat.json if full else str(node.cat)
                return res
            else:
                return {
                    'type': node.op_string,
                    'cat': node.cat.json if full else str(node.cat),
                    'children': [rec(child) for child in node.children]
                }

        if tokens is None:
            tokens = [Token.from_word(word) for word in self.word.split(' ')]
        tokens = list(tokens)
        return rec(self)

    def prolog(self):
        cdef string res = Prolog(self.node_).Get()
        return res.decode("utf-8")

    def ja(self):
        cdef string res = JaCCG(self.node_).Get()
        return res.decode("utf-8")

    def conll(self):
        cdef string res = CoNLL(self.node_).Get()
        return res.decode("utf-8")




from typing import Dict
import numpy as np
cimport numpy as np
import logging
import json
from .cat cimport Category
from libcpp.pair cimport pair


logger = logging.getLogger(__name__)


def normalize(word):
    if word == "-LRB-":
        return "("
    elif word == "-RRB-":
        return ")"
    elif word == "-LCB-":
        return "{"
    elif word == "-RCB-":
        return "}"
    elif word == "-LSB-":
        return "["
    elif word == "-RSB-":
        return "]"
    else:
        return word


def denormalize(word):
    if word == "(":
        return "-LRB-"
    elif word == ")":
        return "-RRB-"
    elif word == "{":
        return "-LCB-"
    elif word == "}":
        return "-RCB-"
    elif word == "[":
        return "-LSB-"
    elif word == "]":
        return "-RSB-"
    word = word.replace(">", "-RAB-")
    word = word.replace("<", "-LAB-")
    return word


def read_pretrained_embeddings(filepath: str) -> np.ndarray:
    nvocab = 0
    io = open(filepath)
    dim = len(io.readline().split())
    io.seek(0)
    for _ in io:
        nvocab += 1
    io.seek(0)
    res = np.empty((nvocab, dim), dtype=np.float32)
    for i, line in enumerate(io):
        line = line.strip()
        if len(line) == 0: continue
        res[i] = line.split()
    io.close()
    return res


def read_model_defs(filepath: str) -> Dict[str, int]:
    res = {}
    for i, line in enumerate(open(filepath, encoding='utf-8')):
        word, _ = line.strip().split(' ')
        res[word] = i
    return res


def remove_comment(line):
    comment = line.find('#')
    if comment != -1:
        line = line[:comment]
    return line.strip()


cdef vector[Cat] cat_list_to_vector(list cats):
    cdef vector[Cat] results
    cdef Category cat
    for cat in cats:
        results.push_back(cat.cat_)
    return results


cdef unordered_set[Cat] cat_list_to_unordered_set(list cats):
    cdef unordered_set[Cat] results
    cdef Category cat
    for cat in cats:
        results.insert(cat.cat_)
    return results


cdef unordered_map[string, unordered_set[Cat]] convert_cat_dict(dict cat_dict):
    cdef unordered_map[string, unordered_set[Cat]] results
    cdef str py_word
    cdef string c_word
    cdef list cats
    for py_word, cats in cat_dict.items():
        c_word = py_word.encode('utf-8')
        results[c_word] = cat_list_to_unordered_set(cats)
    return results


cdef unordered_map[Cat, unordered_set[Cat]] convert_unary_rules(list unary_rules):
    cdef unordered_map[Cat, unordered_set[Cat]] results
    cdef unordered_set[Cat] tmp
    cdef Category cat1, cat2
    for cat1, cat2 in unary_rules:
        if results.count(cat1.cat_) == 0:
            results[cat1.cat_] = unordered_set[Cat]()
        results[cat1.cat_].insert(cat2.cat_)
    return results


cpdef read_unary_rules(filename):
    results = []
    for line in open(filename, encoding='utf-8'):
        line = remove_comment(line.strip())
        if len(line) == 0:
            continue
        cat1, cat2 = line.split()
        cat1 = Category.parse(cat1)
        cat2 = Category.parse(cat2)
        results.append((cat1, cat2))
    logger.info(f'load {len(results)} unary rules')
    return results


cpdef read_cat_dict(filename):
    results = {}
    for line in open(filename, encoding='utf-8'):
        word, *cats = line.strip().split()
        results[word] = [Category.parse(cat) for cat in cats]
    logger.info(f'load {len(results)} cat dictionary entries')
    return results


cpdef read_cat_list(filename):
    results = []
    for line in open(filename, encoding='utf-8'):
        line = remove_comment(line.strip())
        if len(line) == 0:
            continue
        cat = line.split()[0]
        results.append(Category.parse(cat))
    logger.info(f'load {len(results)} categories')
    return results


cpdef read_seen_rules(filename, preprocess):
    cdef list results = []
    cdef Category cat1, cat2
    for line in open(filename, encoding='utf-8'):
        line = remove_comment(line.strip())
        if len(line) == 0:
            continue
        tmp1, tmp2 = line.split()
        cat1 = preprocess(Category.parse(tmp1))
        cat2 = preprocess(Category.parse(tmp2))
        results.append((cat1, cat2))
    logger.info(f'load {len(results)} seen rules')
    return results


cdef unordered_set[CatPair] convert_seen_rules(seen_rule_list):
    cdef unordered_set[CatPair] results
    cdef Category cat1, cat2
    for cat1, cat2 in seen_rule_list:
        results.insert(CatPair(cat1.cat_, cat2.cat_))
    return results


cdef unordered_set[Cat] read_possible_root_categories(list cats):
    cdef unordered_set[Cat] res
    cdef Category tmp
    for cat in cats:
        tmp = Category.parse(cat)
        res.insert(tmp.cat_)
    return res


def read_partial_tree(string):
    stack = []
    spans = []
    words = []
    buf = list(reversed(string.split()))
    counter = 0
    while buf:
        item = buf.pop()
        if item.startswith('<'):
            cat = item[1:]
            cat = None if cat == 'X' else Category.parse(cat)
            stack.append(cat)
            stack.append(counter)
        elif item == '>':
            start = stack.pop()
            cat = stack.pop()
            spans.append((cat, start, counter - start))
        else:
            items = item.split('|')
            if len(items) == 1:
                words.append(items[0])
            elif len(items) == 2:
                cat, word = items
                words.append(word)
                spans.append((Category.parse(cat), counter))
            counter += 1
    assert len(stack) == 0, 'failed to parse partially annotated sentence.'
    return words, spans


def maybe_split_and_join(string):
    if isinstance(string, list):
        split = string
        join = ' '.join(string)
    else:
        assert isinstance(string, str)
        split = string.split(' ')
        join = string
    return split, join


def read_weights(filename, file_type='json'):
    assert file_type == 'json'
    categories = None
    probs = []
    constraints = []
    for line in open(filename):
        json_dict = json.loads(line.strip())
        if categories is None:
            categories = [Category.parse(cat) for cat in json_dict['categories']]

        sent = ' '.join(denormalize(word) for word in json_dict['words'].split(' '))
        dep = np.array(json_dict['heads']).reshape(json_dict['heads_shape']).astype(np.float32)
        tag = np.array(json_dict['head_tags']).reshape(json_dict['head_tags_shape']).astype(np.float32)
        probs.append((tag, dep))
    return probs, categories

